from .petit_ts import Type, TSTypeStore, pseudo_classes
from .base_handler import ClassHandler, BasicHandler, BaseHandler
from .named_types import Named
from .type_spoofer import patch_get_origin_for_Union, spoofer





import ast
import inspect
from typing import Any, Optional

import executing

NAME_TOKEN = '__petit_name__'


class AstFailure(Exception):
    ...


# TODO: make a custom type that reflects the added functionnalities
ExtendedBasicTypeHint = Any


def get_extended_name(item: ExtendedBasicTypeHint) -> Optional[str]:
    if hasattr(item, NAME_TOKEN):
        return item.__dict__[NAME_TOKEN]
    else:
        return None


def set_extended_name(item: ExtendedBasicTypeHint, name: Optional[str]) -> None:
    if name is not None:
        item.__dict__[NAME_TOKEN] = name


def get_parent_assign(node: ast.AST) -> ast.Assign:
    """Look for an ast.Assign node in the parents"""
    while hasattr(node, 'parent'):
        node = node.parent

        if isinstance(node, ast.Assign):
            return node
    raise Exception(
        'Failed to retrieve the variable name.'
    )


def get_variable_name(with_raise: bool = True) -> Optional[str]:
    # TODO: maybe should handle depth, number of f_back
    frame = inspect.currentframe().f_back.f_back
    node = executing.Source.executing(frame).node
    try:
        assignment = get_parent_assign(node)
    except:
        return None
    if len(assignment.targets) > 1:
        if with_raise:
            raise AstFailure(
                f"Too many assignments is not supported {ast.dump(assignment)}")
        return None
    target = assignment.targets[0]
    if isinstance(target, ast.Name):
        return target.id
    if isinstance(target, ast.Attribute):
        return target.attr
    if with_raise:
        raise AstFailure(
            f"Can only get name of a variable or attribute, not {ast.dump(assignment)}"
        )
    return None



from __future__ import annotations

from abc import ABC, abstractstaticmethod
from typing import Generic, TYPE_CHECKING, Any, Dict, List, Optional, Tuple, TypeVar, Union

if TYPE_CHECKING:
    from .petit_ts import TSTypeStore

T = TypeVar('T')


class BaseHandler(ABC, Generic[T]):

    @abstractstaticmethod
    def should_handle(cls: Any, store: TSTypeStore,
                      origin: Optional[type], args: List[Any]) -> bool: ...

    @abstractstaticmethod
    def build(cls: T, store: TSTypeStore, origin: Optional[type],
              args: List[Any]) -> Tuple[Optional[str], Union[str, Dict[str, Any]]]: ...


class BasicHandler(BaseHandler[T]):
    @abstractstaticmethod
    def build(cls: T, store: TSTypeStore, origin: Optional[type],
              args: List[Any]) -> Tuple[Optional[str], str]: ...


class ClassHandler(BaseHandler[T]):
    @abstractstaticmethod
    def is_mapping() -> bool: ...

    @staticmethod
    def is_inline(cls: type) -> bool:
        return False

    @abstractstaticmethod
    def build(cls: T, store: TSTypeStore, origin: Optional[type],
              args: List[Any]) -> Tuple[Optional[str], Union[str, Dict[str, Any]]]: ...



from __future__ import annotations

import inspect
from dataclasses import make_dataclass
from petit_ts.ast_utils import get_extended_name
from typing import (Any, Dict, List, Optional, Set, TypeVar, Union, get_args,
                    get_origin)
from typing import Type as RealType

from .base_handler import BasicHandler, ClassHandler
from .const import DEFAULT_TYPES, INLINE_TOKEN, pseudo_classes
from .handlers import (DataclassHandler, EnumHandler, LiteralHandler,
                       UnionHandler, TupleHandler)
from .utils import SafeCounter, is_array, is_generic, is_mapping, is_optional

# we use this to create different names for the inlined types
"""Be careful with this, it's a global for the library
"""
global_counter = SafeCounter()

BASIC_TYPES = Union[int, str, float, bool, None]


store_hash_function = str


def Type(**kwargs) -> type:
    """Creates a dataclass in order to make an inline type

    example: Type(username=str, password=str) -> 
    ```python
    @dataclass
    class XXX:
        username: str
        password: str
    ```

    It can be used like that :
    ```python
    def handle_input(data: Type(username=str, password=str)) -> Type(error=bool):
        return {'error':False}
    ```
    """
    name = f'{INLINE_TOKEN}{global_counter.increment()}'
    return make_dataclass(name, [(key, type_) for key, type_ in kwargs.items()])


class TSTypeStore:
    """Required in order to link all the types between themselves
    Supports: 
    - `str`, `bool`, `int`, `float`, `Any`, `None` 
    - dict, list
    - Dict[K, V], L[T]
    - @dataclass and generic classes
    - `pydantic.BaseModel`


    (nested types are automaticcaly recurvely added so no need to add in the correct order, 
    generic classes are automaticcally added, if used once), 

    After adding all different wanted types :
    - store.render() and store.get_all_not_inlined()
    - Then you can use store.get_repr() for your classes 
    """

    def __init__(self):
        self.types: Dict[str, TypeStruct] = {}
        # we put them here, because they are less standard than the other
        self.basic_handlers: Set[RealType[BasicHandler]] = {
            UnionHandler,
            LiteralHandler,
            TupleHandler,
        }
        self.class_handlers: Set[RealType[ClassHandler]] = {
            EnumHandler,
            DataclassHandler,
        }
        self.__init_default_type()

    def __init_default_type(self):
        self.types = {
            key: TypeStruct(value, self, default=True) for key, value in DEFAULT_TYPES.items()
        }

    def add_type(self, cls: pseudo_classes) -> None:
        """Adds a type to the store in order to build it's representation in function of the others
        """
        if str(cls) not in self.types:  # check if already built
            self.types[str(cls)] = TypeStruct(cls, self)

    def render_types(self) -> None:
        """Use this to render actual store, not actually required, 
        because get_repr() will render if required
        """
        for type_ in list(self.types.values()):
            type_._render()

    def get_repr(self, cls: pseudo_classes) -> str:
        """Returns the typescript representation of a given type
        """
        # handling generics
        if isinstance(cls, TypeVar):
            return cls.__name__

        if store_hash_function(cls) not in self.types:
            self.add_type(cls)

        return self.types[store_hash_function(cls)].get_repr()

    def get_full_repr(self, cls:pseudo_classes) -> str:
        if isinstance(cls, TypeVar):
            return cls.__name__

        if store_hash_function(cls) not in self.types:
            self.add_type(cls)

        return self.types[store_hash_function(cls)].get_full_repr()

    def get_all_not_inlined(self) -> str:
        """return all the function where a body has to be added to the file
        """
        # ensure rendered
        self.render_types()
        return '\n'.join(
            i.get_full_repr() for i in self.types.values() if i.name is not None
        )

    def add_basic_handler(self, handler: RealType[BasicHandler]) -> None:
        """Adds a `BasicHandler` to the store, in order to add support for a custom class

        if you want to add the support for datetime for example, it's here
        """
        self.basic_handlers.add(handler)

    def add_class_handler(self, handler: RealType[ClassHandler]) -> None:
        """Adds a `ClassHandler` to the store, in order to add support for a custom class
        """
        self.class_handlers.add(handler)

    def add_basic_cast(self, type1: Any, type2: BASIC_TYPES) -> None:
        """For example if you want to cast datetime.datetime directly as str

        will only work for basic types | flat types
        """
        self.types[store_hash_function(type1)] = self.types[store_hash_function(type2)]


class TypeStruct:
    """Internal object used to store the data in order to build, it's typescript representation
    """

    def __init__(self, value: pseudo_classes, store: TSTypeStore, *, default: bool = False):
        self.rendered: bool = False
        self.__repr: str = f"any /* {value} */"
        
        if default:
            self.rendered = True
            self.__repr = value
        
        self.value: str = value
        self.rendering: bool = default
        self.store: TSTypeStore = store
        self.name: Optional[str] = None

    def _make_inline(self, fields: Dict[str, Any]):
        s = []
        for key, type_ in fields.items():
            optional, args = is_optional(type_)
            if optional:
                if len(args) == 2:
                    self.store.add_type(type_)
                    s.append(
                        f'\t{key}?: {self.store.get_repr(args[0])};'
                    )
                # means that we have an Optional[Union[...]]
                else:
                    s.append(
                        f'\t{key}?: {self.store.get_repr(type_)};'
                    )
            else:
                s.append(f'\t{key}: {self.store.get_repr(type_)}')
        self.__repr = '{ ' + ', '.join(s) + ' }'

    def _make_not_inline(self, name: str, fields: Dict[str, Any]):
        self.name = name
        is_generic_, names = is_generic(self.value)
        s: List[str] = []
        if is_generic_:
            s.append(
                f'type {self.name}<{", ".join(self.store.get_repr(n) for n in names)}> = {{'
            )
        else:
            s.append(f'type {self.name}  = {{')
        for key, type_ in fields.items():
            optional, args = is_optional(type_)
            if optional:
                if len(args) == 2:
                    self.store.add_type(type_)
                    s.append(
                        f'\t{key}?: {self.store.get_repr(args[0])};'
                    )
                # means that we have an Optional[Union[...]]
                else:
                    s.append(
                        f'\t{key}?: {self.store.get_repr(type_)};'
                    )
            else:
                s.append(f'\t{key}: {self.store.get_repr(type_)};')
        s.append('};')
        self.__repr = '\n'.join(s)

    def _render(self) -> None:
        """Here is the actual magic :) """
        if self.rendering:
            return
        self.rendering = True

        origin = get_origin(self.value)
        args = get_args(self.value)

        if inspect.isclass(self.value):
            for handler in self.store.class_handlers:
                if handler.should_handle(self.value, self.store, origin, args):
                    name, result = handler.build(
                        self.value, self.store, origin, args)
                    if not handler.is_mapping():
                        self.__repr = result
                    else:
                        if name is None:
                            self._make_inline(result)
                        else:
                            self._make_not_inline(name, result)
                    # we have the check twice, because if you are not a
                    #  mapping, you can still not be inline
                    if name is not None:
                        self.name = name
                    self.rendered = True
                    self.rendering = False
                    return
            # TODO: better handle error
            print(f'Type not Supported {self.value}')
        else:
            for handler in self.store.basic_handlers:
                if handler.should_handle(self.value, self.store, origin, args):
                    name, result = handler.build(
                        self.value, self.store, origin, args
                    )
                    self.__repr = result
                    if name is not None:
                        self.name = name
                    self.rendered = True
                    self.rendering = False
                    return

            if isinstance(self.value, dict):
                s = []
                for key, value in self.value.items():
                    # should check optional
                    s.append(f'{key}: {self.store.get_repr(value)}')
                self.__repr = '{ '+', '.join(s)+' }'
                self.rendered = True
                self.rendering = False
            elif is_array(origin, args):
                type_ = args[0]
                # can't have optional here
                self.__repr = f'{self.store.get_repr(type_)}[]'
                self.rendered = True
                self.rendering = False
            elif is_mapping(origin, args):
                key_type, value_type = args
                # can't have optional here
                self.__repr = f'{{ [key: {self.store.get_repr(key_type)}]: {self.store.get_repr(value_type)} }}'
                self.rendered = True
                self.rendering = False
            # handle generic classes
            elif len(args) > 0:
                self.__repr = f"{self.store.get_repr(origin)}<{', '.join(self.store.get_repr(i) for i in args)}>"
                self.rendered = True
                self.rendering = False
            else:
                print(f'No handler found for this type {self.value}')
                self.rendering = False

    def get_repr(self) -> str:
        """returns the default representation for the type

        - inlined type will return their body

        - not inlined type will return their name
        """
        if not self.rendered:
            self._render()
        if self.name is not None:
            return self.name
        return self.__repr

    def get_full_repr(self) -> str:
        """Used to get the full body of not inlined type
        """
        if not self.rendered:
            self._render()
        if self.name != '':
            return self.__repr



from __future__ import annotations

from dataclasses import is_dataclass
from enum import Enum
from typing import (TYPE_CHECKING, Any, Dict, List, Literal, Optional, Tuple,
                    Union, get_type_hints)

from petit_ts.named_types import NamedLiteral, NamedUnion

from .ast_utils import get_extended_name
from .base_handler import BasicHandler, ClassHandler
from .const import INLINE_TOKEN, NoneType

if TYPE_CHECKING:
    from .petit_ts import TSTypeStore


class UnionHandler(BasicHandler):
    @staticmethod
    def should_handle(cls, store, origin, args) -> bool:
        return origin in (NamedUnion, Union)

    @staticmethod
    def build(cls: Union[Any], store, origin, args) -> Tuple[Optional[str], str]:
        # Union[Any] because Union is like Never

        if (name := get_extended_name(cls)) is None:
            return None, f' | '.join(store.get_repr(arg) for arg in args if arg is not NoneType)
        else:
            return name, f'type {name} = '+' | '.join(store.get_repr(arg) for arg in args) + ';'


class LiteralHandler(BasicHandler):
    @staticmethod
    def should_handle(cls, store, origin, args) -> bool:
        return origin in (Literal, NamedLiteral)

    @staticmethod
    def build(cls: Literal, store, origin, args) -> Tuple[Optional[str], str]:
        name = get_extended_name(cls)
        is_inline = name is None
        s = []
        for arg in args:
            if isinstance(arg, int):
                s.append(f'{arg}')
            elif isinstance(arg, str):
                s.append(f'"{arg}"')
            else:
                # TODO: should add warning logger here
                print('error', arg)
        res = ' | '.join(s)
        if is_inline:
            return None, res
        else:
            return name, f'type {name} = {res};'


class EnumHandler(ClassHandler[Enum]):
    @staticmethod
    def is_mapping() -> bool:
        return False

    @staticmethod
    def should_handle(cls: type, store, origin, args) -> bool:
        return issubclass(cls, Enum)

    @staticmethod
    def build(cls: Enum, store, origin, args) -> Tuple[Optional[str], str]:
        s = []
        name = cls.__name__
        for i in cls:
            if isinstance(i.value, str):
                s.append(f'\t{i.name} = "{i.value}",')
            elif isinstance(i.value, int):
                s.append(f'\t{i.name} = {i.value},')
            else:
                # TODO: handle error, invalid type
                pass
        res = f'enum {name} {{\n' + '\n'.join(s) + '\n};'
        return name, res


class DataclassHandler(ClassHandler):
    @staticmethod
    def is_mapping() -> bool:
        return True

    @staticmethod
    def should_handle(cls: type, store, origin, args) -> bool:
        return is_dataclass(cls)

    def is_inline(cls: type) -> bool:
        return cls.__name__.startswith(INLINE_TOKEN)

    @staticmethod
    def build(cls: type, store, origin, args) -> Tuple[Optional[str], Dict[str, Any]]:
        name = None if cls.__name__.startswith(INLINE_TOKEN) else cls.__name__
        fields = get_type_hints(cls)
        return name, fields


class TupleHandler(BasicHandler):
    @staticmethod
    def should_handle(cls: Any, store: TSTypeStore, origin: Optional[type], args: List[Any]) -> bool:
        return origin is tuple

    def build(cls: Any, store: TSTypeStore, origin: Optional[type], args: List[Any]) -> Tuple[Optional[str], Union[str, Dict[str, Any]]]:
        # Union[Any] because Union is like Never
        if (name := get_extended_name(cls)) is None:
            return None, '[' + f', '.join(store.get_repr(arg) for arg in args if arg is not NoneType) + ']'
        else:
            return name, f'type {name} = ['+', '.join(store.get_repr(arg) for arg in args) + '];'



import typing
from typing import Any, Callable, List, Tuple, Union

from .named_types import NamedUnion


def spoofer(spoofed: List[Tuple[Any, Any]]) -> Callable[[], None]:
    """Will return a function that will **replace** and **patch** the original function `get_origin` from
    the `typing` library.

    Warning, it could cause bug and side effects
    """
    def f() -> None:
        # this functions is copied/pasted from the typing library
        # it's also modified for spoofing types
        def get_origin(tp):
            """Get the unsubscripted version of a type.

            This supports generic types, Callable, Tuple, Union, Literal, Final and ClassVar.
            Return None for unsupported types. Examples::

                get_origin(Literal[42]) is Literal
                get_origin(int) is None
                get_origin(ClassVar[int]) is ClassVar
                get_origin(Generic) is Generic
                get_origin(Generic[T]) is Generic
                get_origin(Union[T, int]) is Union
                get_origin(List[Tuple[T, T]][int]) == list
            """
            if isinstance(tp, typing._GenericAlias):
                for fake, real in spoofed:
                    if tp.__origin__ is fake:
                        return real
                return tp.__origin__
            if tp is typing.Generic:
                return typing.Generic
            return None

        typing.get_origin = get_origin
    return f


"""Will spooef NamedUnion as Union for other framewokr to not have errors
"""
patch_get_origin_for_Union = spoofer([
    (NamedUnion, Union),
])
